{-|
Module      : Data.Graph.Clustering.FLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

# Louvain Algorithm

Our algorithm is divided into two phases that are repeated iteratively.
Assume that we start with a weighted network of N nodes. First, we
assign a different community to each node of the network. So, in this
initial partition there are as many communities as there are nodes.
Then, for each node i we consider the neighbours j of i and we evaluate
the gain of modularity that would take place by removing i from its
community and by placing it in the community of j. The node i is then
placed in the community for which this gain is maximum (in the case of
a tie we use a breaking rule), but only if this gain is positive. If
no positive gain is possible, i stays in its original community. This
process is applied repeatedly and sequentially for all nodes until
no further improvement can be achieved and the first phase is then
complete.

Reference:
Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud;
Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities
in large networks". Journal of Statistical Mechanics: Theory and
Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible.
doi:10.1088/1742-5468/2008/10/P10008.

-}

{-# LANGUAGE ConstrainedClassMethods #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE InstanceSigs      #-}
{-# LANGUAGE NoImplicitPrelude #-}

module Data.Graph.Clustering.FLouvain
  where

import Protolude
import Data.Graph.Inductive
import qualified Data.List as DL

--  "glue" : function to gather/merge communities
--  "klue" : function to split communities
data ClusteringMethod = Glue | Klue
  deriving (Eq)


------------------------------------------------------------------------
-- | Specific FGL needed functions

-- | Get label of an 'LNode'
llab :: LNode a -> a
llab (_, a) = a

-- | Given a 'DynGraph', replace a given 'LNode a' with new label (of type 'a')
replaceLNode ::  (DynGraph gr) => gr a b -> LNode a -> gr a b
replaceLNode gr (n, ln) = gmap replacer gr
  where
    replacer (p, v, l, s) =
      if v == n then (p, v, ln, s) else (p, v, l, s)

-- | Find LNode of a node (i.e. a node with label)
lnode :: (Graph gr) => gr a b -> Node -> Maybe (LNode a)
lnode cgr n = case lab cgr n of
  Nothing -> Nothing
  Just l  -> Just (n, l)

-- | Fold over graph definitions: type and function

-- | Fold over graph type
type CFunFold a b c = Context a b -> c -> c

-- | Fold over graph function
xdfsFoldWith :: (Graph gr)
    => CFun     a b [Node]
    -> CFunFold a b c
    -> c
    -> [Node]
    -> gr a b
    -> c
xdfsFoldWith _ _ acc [] _            = acc
xdfsFoldWith _ _ acc _ g | isEmpty g = acc
xdfsFoldWith d f acc (v:vs) g =
  case match v g of
    (Just c, g')  -> xdfsFoldWith d f (f c acc) (d c++vs) g'
    (Nothing, g') -> xdfsFoldWith d f acc vs g'

------------------------------------------------------------------------

-- Our basic graph. Nodes have custom labels. Edges have weight assigned to them.
newtype Weight = Weight { unWeight :: Double }
  deriving (Show, Eq, Ord)
type FEdge b = (Weight, b)
fedgeWeight :: FEdge b -> Double
fedgeWeight = unWeight . fst
type FGraph a b = Gr a (FEdge b)

-- Used for k_i in formula (2)
newtype NodeWeightSum = NodeWeightSum { unNodeWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- Used for k_i,in in formula (2)
newtype NodeComWeightSum = NodeComWeightSum { unNodeComWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- Probably this structure is better to reduce the number of computations
-- (precompute sum of node weights, which is the k_i variable in formula (2)).
-- type FNode a = (NodeWeightSum, a)
-- fnodeWeightSum :: FNode a -> Double
-- fnodeWeightSum = unNodeWeightSum . fst

-- | This is the m variable in formula (2) of the Louvain paper
newtype GraphWeightSum = GraphWeightSum { unGraphWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_in in formula (2) of the Louvain paper
newtype InWeightSum = InWeightSum { unInWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_tot in formula (2) of the Louvain paper
newtype TotWeightSum = TotWeightSum { unTotWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | Computed Delta_Q value in (2)
newtype DeltaQ = DeltaQ { unDeltaQ :: Double }
  deriving (Show, Eq, Ord)

-- | Type for the clusters we will be creating.
newtype Community = Community { unCommunity ::  ([Node], InWeightSum, TotWeightSum) }
  deriving (Show, Eq, Ord)
comNodes :: Community -> [Node]
comNodes (Community (ns, _, _)) = ns
comInWeightSum :: Community -> InWeightSum
comInWeightSum (Community (_, inWeightSum, _)) = inWeightSum
comTotWeightSum :: Community -> TotWeightSum
comTotWeightSum (Community (_, _, totWeightSum)) = totWeightSum

type CGrNode = Node
type CGrEdge = (InWeightSum, TotWeightSum)

type CGr = Gr Community ()

graphWeight :: FGraph a b -> GraphWeightSum
graphWeight gr = GraphWeightSum $ ufold weight' 0 gr
  where
    weight' (p, _, _, s) acc = acc + sum (map (fedgeWeight . fst) $ p <> s)

-- | Compute initial 'CGr' for a given 'FGraph a b'. This means, put each node
-- in a separate community.
initialCGr :: FGraph a b -> CGr
initialCGr gr = gmap singletonCom gr
  where
    -- A singleton community is given:
    -- the same node id for a community
    -- same incoming/outgoing edges
    -- custom Community label
    singletonCom (p, v, l, s) = (p', v, Community ([v], iws, tws), s')
      where
        p' = map edgeComRemap p
        s' = map edgeComRemap s
        edgeComRemap (_, n) = ((), n)
        edges = lneighbors gr v
        -- no internal links
        iws = InWeightSum 0.0
        -- just sum over the edges coming into/out of node v
        tws = TotWeightSum $ sum $ map (fedgeWeight . fst) edges

-- ALGORITHM

-- | Q function from Louvain paper (1).
modularity :: Gr a b -> CGr -> Double
modularity gr cgr = 0.0

type Delta = Community -> NodeWeightSum -> NodeComWeightSum -> GraphWeightSum -> DeltaQ
-- | Delta Q function from Louvain paper (2).
delta :: Delta
delta com ki kiin m = DeltaQ $ acc - dec
  where
    inWeightSum = comInWeightSum com
    totWeightSum = comTotWeightSum com
    acc = accL - accR*accR
    accL = 0.5 * (unInWeightSum inWeightSum + 2.0 * (unNodeComWeightSum kiin)) / (unGraphWeightSum m)
    accR = 0.5 * (unTotWeightSum totWeightSum + unNodeWeightSum ki) / (unGraphWeightSum m)
    dec = decL - decM * decM - decR * decR
    decL = 0.5 * (unInWeightSum inWeightSum) / (unGraphWeightSum m)
    decM = 0.5 * (unTotWeightSum totWeightSum) / (unGraphWeightSum m)
    decR = 0.5 * (unNodeWeightSum ki) / (unGraphWeightSum m)

-- | One iteration step takes the graph and existing communities as a graph and
-- computes new community graph

-- NOTE: xdfsFoldWith only iterates with the defined Context. In each step of
-- the algorithm, we need full node information however (i.e. all edges
-- connected to a node), so instead of calling just 'step' we are forced to call via:
-- 'step . context gr . node''
-- We could avoid the higher complexity, eg. by precomputing the whole graph
-- into a HashMap Node [Edge].
iteration :: FGraph a b -> CGr -> CGr
iteration gr cs = xdfsFoldWith suc' (\(_, v, _, _) -> step gw $ context gr $ v) cs (nodes gr) gr
  where
    gw = graphWeight gr
    --weightSum = ufold weightSum' 0 gr
    --weightSum' (p, v, l, s) acc = acc + (sum )

-- TODO Remember to filter out empty Communities
-- | Step for one node. We try re-assign it to a neighbouring community, where
-- the increase of modularity for graph will be the largest
step :: GraphWeightSum -> CFunFold a (FEdge b) CGr
step gw ctx@(p, v, l, s) cgr = newCgr
  where
    newCgr = case mNc of
      Nothing -> cgr
      Just nc ->
        if bestFitdq > 0.0 then
          let newBestFitCom = makeMove Into bestFitCom
              newNc = makeMove OutOf nc
              in
            replaceLNode (replaceLNode cgr newNc) newBestFitCom
        else
          cgr

    (bestFitCom, DeltaQ bestFitdq) = maximumBy (\(_, deltaq1) (_, deltaq2) -> compare deltaq1 deltaq2) deltas

    mNc :: Maybe (LNode Community)
    mNc = nodeCommunity v cgr
    ncs :: [LNode Community]
    ncs = nodeNeighbours v cgr

    -- We move node from community nc into ncs
    -- moves :: Maybe (LNode Community, [LNode Community])
    -- moves = case mNc of
    --   Nothing -> Nothing
    --   Just nc -> Just ( makeMove OutOf nc
    --                   , map (makeMove Into) ncs )

    makeMove :: Direction -> LNode Community -> LNode Community
    makeMove direction (cn, c) = (cn, moveNodeWithNeighbours (p <> s) v direction c)

    -- k_i variable in formula (2)
    ki :: NodeWeightSum
    ki = NodeWeightSum $ sum $ map (fedgeWeight . fst) $ p <> s

    deltas :: [(LNode Community, DeltaQ)]
    deltas = map (\c -> (c, delta' c)) ncs

    delta' :: LNode Community -> DeltaQ
    delta' com = delta (llab com) ki kiin gw
      where
        -- k_i,in variable in formula (2)
        kiin :: NodeComWeightSum
        kiin = nodeComWeightSum (llab com) ctx

    -- TODO Compute \Delta Q (gain of moving node v into Community C) which consists of:
    -- - Community InWeightSum
    -- - Community TotWeightSum
    -- - sum of weights of links incident to node v (taken from 'p' and 's')
    -- - sum of weights of links from node v to nodes in C (taken from 'p' and 's')
    -- - sum of weights of all the links in the network
    -- So the Delta function takes as parameters:
    -- C, (edges from C in cgr), (edges from v in gr), (edges from v to C), (edges in gr)

nodeComWeightSum :: Community -> Context a (FEdge b) -> NodeComWeightSum
nodeComWeightSum com (_, _, _, s) =
  NodeComWeightSum $ sum $ map (fedgeWeight . fst) $ filter (\(_, n) -> n `elem` comNodes com) s

-- COMMUNITY GRAPH FUNCTIONS

-- | 'Direction' when moving node 'Into'/'OutOf' community
data Direction = Into | OutOf

-- | Given 'Node' and 'Community' graph, find the 'LNode' of 'Community' which
-- contains the node
nodeCommunity :: Node -> CGr -> Maybe (LNode Community)
nodeCommunity n cgr = head (filter f $ labNodes cgr)
  where
    f :: (a, Community) -> Bool
    f (_, com) = n `elem` comNodes com

-- | Find 'LNode's of 'Community' graph neighbouring a given node
nodeNeighbours :: Node -> CGr -> [LNode Community]
nodeNeighbours n cgr =
  case nodeCommunity n cgr of
    Nothing        -> []
    Just (cn, _) -> mapMaybe (lnode cgr) (neighbors cgr cn)

-- | Find 'Ajd CGrEdge's of 'Community' graph neighbouring a given node
-- nodeLNeighbours :: Node -> CGr -> Adj CGrEdge
-- nodeLNeighbours n cgr =
--   case nodeCommunity n cgr of
--     Nothing        -> []
--     Just (cn, _) -> lneighbors cgr cn

-- | Moves 'Node' in the 'Direction' of 'Community' and recomputes 'Community''s weights
moveNode :: FGraph a b -> Node -> Direction -> Community -> Community
moveNode gr n direction c = moveNodeWithNeighbours lnNeighbors n direction c
  where
    --lnNeighbors :: Adj (FEdge b)
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' above but with only node neighbours, not whole graph
moveNodeWithNeighbours :: Adj (FEdge b) -> Node -> Direction -> Community -> Community
moveNodeWithNeighbours lnNeighbors n direction (Community (ns, inwsum, totwsum)) =
  Community (newNs, newInWsum, newTotWsum)
    where

      newNs = case direction of
        Into  -> n:ns
        OutOf -> DL.delete n ns

      (newInWsum, newTotWsum) = computeWeights direction (inwsum, totwsum) sumN sumNonCom

      -- Update InWeightSum with connections between node and the community
      sumN :: Double
      sumN = sum $ map (fedgeWeight . fst) comNeighbors

      -- Update TotWeightSum, subtracting connections between node and community
      -- and adding connections of node to non-community
      sumNonCom :: Double
      sumNonCom = sum $ map (fedgeWeight . fst) nonComNeighbors

      -- Node Adj Context
      --comNeighbors    :: Adj (FEdge b)
      comNeighbors    = filter (\ln -> snd ln `elem`    ns) lnNeighbors

      --nonComNeighbors :: Adj (FEdge b)
      nonComNeighbors = filter (\ln -> snd ln `notElem` ns) lnNeighbors

-- | Recomputes 'InWeightSum' when node is moved in the 'Direciton' of
-- 'Community'. Given parameters are:
computeWeights :: Direction -> (InWeightSum, TotWeightSum) -> Double -> Double -> (InWeightSum, TotWeightSum)
computeWeights Into (InWeightSum inwsum, TotWeightSum totwsum) sumN sumNonCom =
  (InWeightSum $ inwsum  + sumN, TotWeightSum $ totwsum + sumN - sumNonCom)
computeWeights OutOf (InWeightSum inwsum, TotWeightSum totwsum) sumN sumNonCom =
  (InWeightSum $ inwsum  - sumN, TotWeightSum $ totwsum - sumN + sumNonCom)


{-
-- | Moves 'Node' between two 'Community'ies, recomputing their weights
moveNode' :: FGraph a -> Node -> (Community, Community) -> (Community, Community)
moveNode' gr n cFromTo = moveNodeWithNeighbours' lnNeighbors n cFromTo
  where
    lnNeighbors :: Adj FEdge
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' but with direct neighbours list
moveNodeWithNeighbours' :: Adj FEdge -> Node -> (Community, Community) -> (Community, Community)
moveNodeWithNeighbours' lnNeighbors n ((Community (fromNs, fromInWSum, fromTotWSum)), (Community (toNs, toInWSum, toTotWSum))) =
  (newFrom, newTo)
  where
    newFrom = Community (DL.delete n fromNs, newFromInWSum, newFromTotWSum)
    newTo = Community (n:toNs, newToInWSum, newToTotWSum)

    -- Node is removed, so we reduce the internal weight sum
    newFromInWSum = fromInWSum - sumFromN
    newFromTotWSum = fromTotWSum + sumFromN - sumFromNonN

    -- Node is added, so we increase the internal weight sum
    newToInWSum = toInWSum + sumToN
    newToTotWSum = toTotWSum + sumToN - sumToNonN

    sumFromN :: InWeightSum
    sumFromN = sum $ map fst fromComNeighbors
    sumFromNonN :: TotWeightSum
    sumFromNonN = sum $ map fst fromNonComNeighbors

    sumToN :: InWeightSum
    sumToN = sum $ map fst toComNeighbors
    sumToNonN :: TotWeightSum
    sumToNonN = sum $ map fst toNonComNeighbors

    fromComNeighbors :: Adj FEdge
    fromComNeighbors = filter (\ln -> snd ln `elem` fromNs) lnNeighbors
    fromNonComNeighbors :: Adj FEdge
    fromNonComNeighbors = filter (\ln -> snd ln `notElem` fromNs) lnNeighbors
    toComNeighbors :: Adj FEdge
    toComNeighbors = filter (\ln -> snd ln `elem` toNs) lnNeighbors
    toNonComNeighbors :: Adj FEdge
    toNonComNeighbors = filter (\ln -> snd ln `notElem` toNs) lnNeighbors
-}
